# Copyright (c) 2024-2025 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.
"""Tools for starting up interactive sessions."""

import os

from IPython.core.extensions import ExtensionManager

from ..mpi import MPI
from ..utils import Environment, Logger


def start_parallel(procs=1, threads=1, nice=True, auto_mpi=False, shell=None):
    """Attempt to start up ipyparallel with mpi4py and OpenMP.

    Args:
        procs (int):  The number of processes to use.
        threads (int):  The number of OpenMP threads per process.
        nice (bool):  If True, "nice" the processes to a lower
            priority.

    Returns:
        (int):  The number of processes started.

    """
    log = Logger.get()
    env = Environment.get()
    
    if procs > 1:
        try:
            import ipyparallel as ipp

            cluster = ipp.Cluster(engines="mpi", n=procs)
            client = cluster.start_and_connect_sync()
            client.block = True
            if nice:
                # Optionally nice the individual processes if running on a
                # shared node.
                if procs > 1:
                    import psutil

                    psutil.Process().nice(
                        20 if psutil.POSIX else psutil.IDLE_PRIORITY_CLASS
                    )
            # Optionally enable automatic use of MPI
            if auto_mpi and shell is not None:
                # Turn on automatic use of MPI
                shell.run_line_magic("autopx")
        except Exception:
            procs = 1
            print("Failed to start ipyparallel cluster, using one process.")
    extmanager = ExtensionManager()
    try:
        _ = extmanager.load_extension("wurlitzer")
    except Exception:
        # Must be running outside IPython shell
        pass

    if procs == 1:
        rank = 0
        comm = None
    else:
        comm = MPI.COMM_WORLD
        rank = comm.rank

    max_threads = env.max_threads()
    if threads > max_threads:
        msg = f"Requested threads per process ({threads}) exceeds"
        msg += f" the maximum ({max_threads}).  Using {max_threads} instead."
        log.warning_rank(msg, comm=comm)
        threads = max_threads
    env.set_threads(threads)

    log.info_rank(f"Using {procs} processes with {threads} threads each.", comm=comm)
